from bolsonaro import LOG_PATH
from bolsonaro.error_handling.logger_factory import LoggerFactory

from sklearn.ensemble import RandomForestRegressor
from sklearn.linear_model import OrthogonalMatchingPursuit
from sklearn.base import BaseEstimator
import numpy as np


class OmpForestRegressor(BaseEstimator):

    DEFAULT_SCORE_METRIC = 'mse'

    def __init__(self, models_parameters):
        self._regressor = RandomForestRegressor(n_estimators=models_parameters.forest_size,
            random_state=models_parameters.seed)
        self._models_parameters = models_parameters
        self._logger = LoggerFactory.create(LOG_PATH, __name__)

    @property
    def forest(self):
        return self._forest

    def set_forest(self, forest):
        self._forest = forest
        self._regressor.estimators_ = forest

    @property
    def weights(self):
        return self._weights

    def set_weights(self, weights):
        self._weights = weights

    @property
    def models_parameters(self):
        return self._models_parameters

    def fit(self, X_forest, y_forest, X_omp, y_omp):
        self._forest = self._train_forest(X_forest, y_forest)
        self._weights = self._extract_subforest(X_omp, y_omp)
        return self

    def score_regressor(self, X, y):
        return self._regressor.score(X, y)

    def predict(self, X):
        """
        Apply the OMPForestRegressor to X.

        :param X:
        :return:
        """
        D = self._forest_prediction(X)

        if self._models_parameters.normalize_D:
            D /= self._forest_norms

        predictions = D @ self.weights

        return predictions

    def score(self, X, y, metric=DEFAULT_SCORE_METRIC):
        """
        Evaluate OMPForestRegressor on (`X`, `y`) using `metric`

        :param X:
        :param y:
        :param metric:
        :return:
        """
        predictions = self.predict(X)

        if metric == 'mse':
            evaluation = np.mean(np.square(predictions - y))
        else:
            raise ValueError("Unsupported metric '{}'.".format(metric))

        return evaluation

    def _train_forest(self, X, y):
        self._regressor.fit(X, y)
        forest = self._regressor.estimators_
        return forest
    
    def _extract_subforest(self, X, y):
        """
        Given an already estimated regressor: apply OMP to get the weight of each tree.

        The X data is used for interrogation of every tree in the forest. The y data
        is used for finding the weights in OMP.

        :param X: (n_sample, n_features) array
        :param y: (n_sample,) array
        :return:
        """
        self._logger.debug("Forest make prediction on X")
        D = self._forest_prediction(X)

        if self._models_parameters.normalize_D:
            # question: maybe consider other kinds of normalization
            self._logger.debug("Compute norm of predicted vectors on X")
            self._forest_norms = np.linalg.norm(D, axis=0)
            D /= self._forest_norms

        omp = OrthogonalMatchingPursuit(
            n_nonzero_coefs=self._models_parameters.extracted_forest_size,
            fit_intercept=False, normalize=False)
        self._logger.debug("Apply orthogonal maching pursuit on forest for {} extracted trees."
                           .format(self._models_parameters.extracted_forest_size))
        omp.fit(D, y)
        weights = omp.coef_
        # question: why not to use directly the omp estimator instead of bypassing it using the coefs?
        return weights

    def _forest_prediction(self, X):
        return np.array([tree.predict(X) for tree in self._forest]).T
